# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Importer script
Can import a function from a given script from file path
"""
import sys
import traceback
import importlib.util


def dynamic_import_class(module_path, class_name):
    """Dynamically imports some class/function from a python file

    Args:
        module_path (str) : path to python file (ex: ./bla/foo.py)
        class_name (str) : name of class/function to import from it

    Returns:
        class_attr (class) : object imported from module
    """
    spec = importlib.util.spec_from_file_location("dynimportmodulename", module_path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    class_attr = getattr(mod, class_name)
    return class_attr


def dynamic_import_module(module_path):
    """Dynamically imports some class/function from a python file

    Args:
        module_path (str) : path to python file (ex: ./bla/foo.py)

    Returns:
        mod, spec
    """
    spec = importlib.util.spec_from_file_location("dynimportmodulename", module_path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return spec, mod


def import_and_test_class(module_path, class_name):
    """Tests importing some class/function from a python file

    Args:
        module_path (str) : path to python file (ex: ./bla/foo.py)
        class_name (str) : name of class/function to import from it

    Returns:
        class_attr (class) : object imported from module
    """
    # WORK IN PROGRESS
    # test if module_path can be found in path
    # test if class_name exists in module_path
    # test return attr if class
    import_success = True
    message = None
    try:
        imported_class = dynamic_import_class(module_path, class_name)
    except:
        import_success = False
        message = traceback.format_exc()

    assert import_success, """
        Importing class '{}' from module path '{}' did not succeed.
        Current python path is {}.
        Traceback from exception:
        {}
        """.format(
        class_name, module_path, sys.path, message
    )

    return imported_class



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import json
from hashlib import sha256


# This list is up-to-date as of 03/23/2022.
# Contact aims-team@microsoft.com if your eyes-on subscription is not included.
EYES_ON_SUBSCRIPTIONS = set(
    [
        "d3259042d8e1f4f5b44ac56c9ca150955cd6a48d22180cf15f37133b850f784e",
        "887f1c5c3f867712fb0cd743e67325643e95a4c29954bce9ebef08922749daf1",
        "9ad297aa5e16962f7943c3343dfe8df27a685bb1987302161d1aa654d60e8f60",
        "74dea0d792de7adb561775eb00898ee41801f467825ce0942607461bcac4419c",
        "dfcaf96e30a340fae2fcd5871a5e94cf3805f6521d5e2571d55b097df857c9c3",
        "89b8bdaf7904e676c93e24be918b83c64fe904a4097821e3def68c0c33d1599b",
        "7f4ce25ded92373dd8ee223d40133e33632601866a7a035c050c13998098bf3b",
        "897107fb585b318ffa494339fd6bd5c38e96548c90c8809d8522c6e71b471cad",
        "e93b407a7a2b559f2cd8ea67d9a1bffbbbca1f616bd47347aea299a993bb4120",
        "71adef91aa2cb4676d995e8b3843dfe75d65e200dd8b552e0c7f5195c043f946",
        "2585e2381d1a9a413392c6c5372ca0bafdb845de17f0bfb5536d340401dee699",
        "7f6d1d5c3d3f46b0459bbb041e6ebebcfceff2e202aa02cb23301761e51b9c0b",
        "d675de4926b9534967be30a8cb035cb3d14b2d02b00ca3077d451a48910841ac",
        "84631d56ce357894baa249912456077e15f899c16639cdaef8c531013990db87",
        "566032778897f87065c0fab68a33e8b184e23a007d392f5a71dbb5f46bf2c460",
        "7141d27136d73133d699a94ebb36d29b524767949c07ba175d265420792b8448",
        "544181f90401adb3bfe691467fdd8dfd37b2eb689b15bcc8585adc6bb81df5d0",
        "9ad143f7e6c66dfa1fa39f147ed258fd584e05792b11a7dd3600537e28e39191",
        "f438a35eb69dd495c8ccb75cf6e554da2ab772ea923b801c822a64b3d43c0926",
        "0e977328631074f17aedfeec10ed487ff71231501357b5eb8b9c454a5604adf0",
        "8776711f970cd8aa00047879e5bee9ba9a0a9a0890241cb5ebd94a7173563bad",
        "d2e95d22a8e429f4e5b6291f3daa0cee13b518241ad30d3f159f9fd242ec7df9",
        "9fe3e3bd8dfdcd3e707c6840161d625460703d8e1e779b8b0e4965f7b1d17d6c",
        "196c596ac5e78ff19b6216eeb4faf8d3b5c9379c88061a64bbfd6da91315ffa7",
        "d194e7433fdc4e28c33fd17cdb42db67fd84d0728f9471ad106b8c4f08fa78ff",
        "770f1203c522f031dde28fab282ee3fbdeb62afa05d3792610254e97171828c5",
    ]
)
EYES_OFF_TENANT_ID = "cdc5aeea-15c5-4db6-b079-fcadd2505dc2"


def get_hashed_eyes_on_subs():
    """
    This method is used to generated EYES_ON_SUBSCRIPTIONS.
    """
    with open("torus_subscription.json", "r") as f:
        # get torus_subscription.json from https://resources.azure.com/subscriptions with torus account  # noqa: E501
        all_info = json.load(f)
    all_sub = []
    for i in all_info["value"]:
        all_sub.append(i["subscriptionId"].lower())

    # Heron known eyes-off subscriptions: PROD MOP sandboxes in the following list
    # https://eng.ms/docs/experiences-devices/m365-core/microsoft-search-assistants-intelligence-msai/substrate-intelligence/ai-training-heron/documentation/subscriptions  # noqa: E501
    eyes_off_sub = [
        "08047947-f71e-4462-a09d-266e3d34c431",
        "60d27411-7736-4355-ac95-ac033929fe9d",
        "2dea9532-70d2-472d-8ebd-6f53149ad551",
    ]

    eyes_on_sub = []
    for i in all_sub:
        if i not in eyes_off_sub:
            eyes_on_sub.append(sha256(i.encode()).hexdigest())
    return eyes_on_sub


def is_eyesoff_helper(tenant_id, subscription_id):
    if tenant_id and tenant_id != EYES_OFF_TENANT_ID:
        # tenant_id could be None for HDI jobs
        return False
    else:
        hashed_subscription_id = sha256(subscription_id.encode()).hexdigest()
        return hashed_subscription_id not in EYES_ON_SUBSCRIPTIONS



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Helper code to run Ray distributed scripts [EXPERIMENTAL]
"""
import os
import logging
import uuid
import time
import socket

import ray

from shrike._core import experimental
from shrike.distributed import EXPERIMENTAL_WARNING_MSG
from shrike.distributed.cluster_auto_setup import ClusterAutoSetupHandler


class RayClusterSetupHandler(ClusterAutoSetupHandler):
    def __init__(self):
        """Constructor"""
        super().__init__()

        # ray init settings
        self.head_address = "auto"
        self.head_port = 6379
        self.redis_password = None

    #################
    # SETUP METHODS #
    #################

    def setup_local(self):
        """Setup method if custom_sync_setup=False"""
        self.logger.info(f"{self.__class__.__name__}.setup_local() called.")
        # nothing to do here

    def setup_head_node(self):
        """Setup to run only on head node"""
        self.logger.info(
            f"{self.__class__.__name__}.setup_head_node() called to set up HEAD node."
        )
        # create setup config
        if "AZ_BATCHAI_NODE_IP" in os.environ:
            local_ip = os.environ["AZ_BATCHAI_NODE_IP"]
            self.logger.info(f"Obtained IP from env varAZ_BATCHAI_NODE_IP={local_ip}")
        else:
            local_hostname = socket.gethostname()
            local_ip = socket.gethostbyname(local_hostname)
            self.logger.info(f"Obtained IP from socket: {local_ip}")

        self.head_address = local_ip
        self.head_port = 6379
        self.redis_password = str(uuid.uuid4())

        # record what's needed to setup cluster nodes
        self.setup_config_add_key("head_address", self.head_address)
        self.setup_config_add_key("head_port", self.head_port)
        self.setup_config_add_key("redis_password", self.redis_password)

        # on head node, init should use "auto"
        self.head_address = "auto"

        # run ray cli
        ray_setup_command = [
            "ray",
            "start",
            "--head",
            f"--port={self.head_port}",
            f"--redis-password={self.redis_password}",
        ]
        self.run_cli_command(ray_setup_command)

    def setup_cluster_node(self):
        """Setup to run only on non-head cluster nodes"""
        self.logger.info(f"{self.__class__.__name__}.setup_cluster_node() called")
        self.head_address = self.setup_config_get_key("head_address")
        self.head_port = self.setup_config_get_key("head_port")
        self.redis_password = self.setup_config_get_key("redis_password")

        # run ray cli
        ray_setup_command = [
            "ray",
            "start",
            f"--address={self.head_address}:{self.head_port}",
            f"--redis-password={self.redis_password}",
        ]
        self.run_cli_command(ray_setup_command)

    def head_node_teardown(self):
        """Un-setup a cluster node"""
        self.logger.info(f"{self.__class__.__name__}.head_node_teardown() called")
        self.run_cli_command(["ray", "stop", "--force", "-v"])

    def cluster_node_teardown(self):
        """Un-setup a cluster node"""
        self.logger.info(f"{self.__class__.__name__}.cluster_node_teardown() called")
        self.run_cli_command(["ray", "stop", "--force", "-v"])


###########################
# USER FRIENDLY FUNCTIONS #
###########################

_SETUP_HANDLER = None


@experimental(EXPERIMENTAL_WARNING_MSG)
def init(address=None, _redis_password=None):
    """User-friendly function to initialize the script using ClusterAutoSetupHandler"""
    global _SETUP_HANDLER
    _SETUP_HANDLER = RayClusterSetupHandler()
    _SETUP_HANDLER.initialize_run()

    multinode_config = _SETUP_HANDLER.multinode_driver.get_multinode_config()
    if multinode_config is None:
        raise Exception(
            "init() needs multinode_driver.get_multinode_config() which is None."
        )
    elif not multinode_config.multinode_available:
        # we're running single node or local
        return ray.init(address=address, _redis_password=_redis_password)
    elif multinode_config.main_node:
        if _SETUP_HANDLER.redis_password is None:
            ray_init_retval = ray.init(address="auto")
        elif isinstance(_SETUP_HANDLER.redis_password, str):
            ray_init_retval = ray.init(
                address="auto", _redis_password=_SETUP_HANDLER.redis_password
            )
        else:
            raise ValueError(
                "Value in redis_password is of wrong type"
                + " {type(_SETUP_HANDLER.redis_password)}"
            )

        # making absolutely sure all nodes are there...
        for _ in range(60):
            logging.getLogger(__name__).info(
                "Waiting for ray cluster to reach available nodes size..."
                + f"[{len(ray.nodes())}/{multinode_config.world_size}]"
            )
            if len(ray.nodes()) >= multinode_config.world_size:
                break
            time.sleep(1)
        else:
            raise Exception(
                "Could not reach maximum number of nodes before 60 seconds."
            )

        return ray_init_retval
    else:
        return None


@experimental(EXPERIMENTAL_WARNING_MSG)
def shutdown():
    """User-friendly function to teardown the script using ClusterAutoSetupHandler"""
    global _SETUP_HANDLER
    if _SETUP_HANDLER is not None:
        _SETUP_HANDLER.finalize_run()

        # use shutdown only if we're on main node
        if _SETUP_HANDLER.multinode_driver.get_multinode_config().main_node:
            ray.shutdown()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Generic process to automatically setup a cluster from within a job.
Leverages MPI to synchronize setup between nodes, and execute
a provided command to start/stop (ex: ray start).

This script is the generic class on which other framework-specific
classes will be derived (ex: Ray, Dask, ...).
"""
import logging
import subprocess
from typing import Any, List
import uuid
import time

from shrike._core import experimental
from shrike.distributed import EXPERIMENTAL_WARNING_MSG
from shrike.distributed.mpi_driver import MultiNodeMPIDriver


class ClusterAutoSetupHandler:
    # NOTE: not even sure we need tags, but let's do it
    COMM_TAG_CLUSTER_SETUP = 42
    COMM_TAG_SETUP_FINISHED = 43
    COMM_TAG_CLUSTER_SHUTDOWN = 44

    def __init__(self, mpi_init_mode: int = 3):
        """Generic initialization for all script classes.

        Args:
            mpi_init_mode (int): mode to initialize MPI (default: THREAD)
        """
        self.logger = logging.getLogger(__name__)

        # keep those for initialization later
        self.multinode_driver = MultiNodeMPIDriver(mpi_init_mode=mpi_init_mode)

        # this will be used to collect cluster config
        self._setup_config = {}

    def run_cli_command(self, cli_command: List[str], timeout: int = 60):
        """Runs subprocess for a cli setup command"""
        self.logger.info(f"Launching cli with command: {cli_command}")
        cli_command_call = subprocess.run(
            cli_command,
            # stdout=PIPE,
            # stderr=PIPE,
            universal_newlines=True,
            check=False,  # will not raise an exception if subprocess fails
            timeout=timeout,  # TODO: more than a minute would be weird?
            # env=custom_env
        )
        self.logger.info(f"return code: {cli_command_call.returncode}")

        if cli_command_call.returncode != 0:
            raise RuntimeError("Cli command returned code != 0")

        return cli_command_call.returncode

    #################
    # SETUP METHODS #
    #################

    def setup_config_add_key(self, key: str, value: Any) -> None:
        self._setup_config[key] = value

    def setup_config_get_key(self, key: str, default_value: Any = None) -> Any:
        return self._setup_config.get(key, default_value)

    # For specific setups, override methods below

    def setup_local(self):
        """Setup method if custom_sync_setup=False"""
        self.logger.info(f"{self.__class__.__name__}.setup_local() called.")

    def setup_head_node(self):
        """Setup to run only on head node"""
        self.logger.info(
            f"{self.__class__.__name__}.setup_head_node() called to set up HEAD node."
        )
        self.setup_config_add_key("_session_id", str(uuid.uuid4()))

    def setup_cluster_node(self):
        """Setup to run only on non-head cluster nodes"""
        self.logger.info(f"{self.__class__.__name__}.setup_cluster_node() called.")

    def head_node_teardown(self):
        """Un-setup a cluster node"""
        self.logger.info(f"{self.__class__.__name__}.head_node_teardown() called.")

    def cluster_node_teardown(self):
        """Un-setup a cluster node"""
        self.logger.info(f"{self.__class__.__name__}.cluster_node_teardown() called.")

    ############
    # MPI COMM #
    ############

    def broadcast_config_from_head_to_cluster_nodes(self):
        """[HEAD only] Sends the cluster setup params to each non-head node"""
        self.logger.info(
            f"Sending cluster setup from head to cluster nodes: {self._setup_config}"
        )
        for i in range(1, self.multinode_driver.get_multinode_config().world_size):
            self.multinode_driver.get_comm().send(
                self._setup_config,
                i,
                tag=ClusterAutoSetupHandler.COMM_TAG_CLUSTER_SETUP,
            )

    def listen_cluster_setup_from_head_node(self):
        """[NODE only] Waits for head node to send cluster setup params"""
        self._setup_config = self.multinode_driver.get_comm().recv(
            source=0, tag=ClusterAutoSetupHandler.COMM_TAG_CLUSTER_SETUP
        )
        self.logger.info(f"Obtained cluster setup from head node: {self._setup_config}")

    def wait_on_nodes_setup_ready(self):
        """[HEAD only] Waits for each node to report completion of their setup"""
        self.logger.info("Checking setup status from each node...")

        # loop on each node in the world and wait for status
        for i in range(1, self.multinode_driver.get_multinode_config().world_size):
            status = self.multinode_driver.get_comm().recv(
                source=i, tag=ClusterAutoSetupHandler.COMM_TAG_SETUP_FINISHED
            )
            self.logger.info(f"Node #{i}: {status}")

            if status != "OK":
                raise RuntimeError(f"Node #{i} failed to setup, status=={status}.")

    def report_node_setup_complete(self):
        """[NODE only] Report to head that this node setup is complete"""
        self.logger.info("Reporting status OK to head node.")
        self.multinode_driver.get_comm().send(
            "OK", 0, tag=ClusterAutoSetupHandler.COMM_TAG_SETUP_FINISHED
        )

    def broadcast_shutdown_signal(self):
        """[HEAD only] Sends message to shutdown to all nodes"""
        for i in range(1, self.multinode_driver.get_multinode_config().world_size):
            self.logger.info(f"Broadcasting shutdown message to node #{i}")
            self.multinode_driver.get_comm().send(
                "SHUTDOWN", i, tag=ClusterAutoSetupHandler.COMM_TAG_CLUSTER_SHUTDOWN
            )

    def non_block_wait_for_shutdown(self):
        """[NODE only] Checks if head node has sent shutdown message"""
        return self.multinode_driver.get_comm().iprobe(
            source=0, tag=ClusterAutoSetupHandler.COMM_TAG_CLUSTER_SHUTDOWN
        )

    ##################
    # SCRIPT METHODS #
    ##################

    def initialize_run(self):
        """Initialize the component run, opens/setups what needs to be"""
        self.logger.info(f"Call to {self.__class__.__name__}.initialize_run()...")

        # initialize mpi comm
        self.multinode_driver.initialize()
        multinode_config = self.multinode_driver.get_multinode_config()

        if multinode_config.multinode_available:
            # initialize setup accross nodes
            if multinode_config.main_node:
                # run setup on head node
                self.setup_head_node()

                # send cluster config to all other nodes
                self.broadcast_config_from_head_to_cluster_nodes()

                # then wait for all nodes to finish setup
                self.wait_on_nodes_setup_ready()
            else:
                # get cluster setup from head node using mpi
                self.listen_cluster_setup_from_head_node()

                # run setup on cluster node
                self.setup_cluster_node()

                # then report that setup is complete
                self.report_node_setup_complete()
        else:
            # run custom method for setup
            self.setup_local()

    def finalize_run(self):
        """Finalize the run, close what needs to be"""
        self.logger.info(f"Call to {self.__class__.__name__}.finalize_run()...")

        multinode_config = self.multinode_driver.get_multinode_config()

        if multinode_config.multinode_available:
            # properly teardown all nodes
            if multinode_config.main_node:
                # run teardown on head node
                self.head_node_teardown()

                # send signal to teardown to each node
                self.broadcast_shutdown_signal()
            else:
                # wait for teardown signal from head node
                while True:
                    self.logger.info("Waiting for teardown signal from HEAD node...")

                    if self.non_block_wait_for_shutdown():
                        break

                    time.sleep(10)

                # run teardown on cluster
                self.cluster_node_teardown()

        # clean exit from driver
        self.multinode_driver.finalize()


###########################
# USER FRIENDLY FUNCTIONS #
###########################

_SETUP_HANDLER = None


@experimental(EXPERIMENTAL_WARNING_MSG)
def init() -> Any:
    """User-friendly function to initialize the script using ClusterAutoSetupHandler"""
    global _SETUP_HANDLER
    _SETUP_HANDLER = ClusterAutoSetupHandler()
    _SETUP_HANDLER.initialize_run()

    if _SETUP_HANDLER.multinode_driver.get_multinode_config().main_node:
        return _SETUP_HANDLER.multinode_driver.get_multinode_config()
    else:
        return None


@experimental(EXPERIMENTAL_WARNING_MSG)
def shutdown():
    """User-friendly function to teardown the script using ClusterAutoSetupHandler"""
    global _SETUP_HANDLER
    if _SETUP_HANDLER is not None:
        _SETUP_HANDLER.finalize_run()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Pipeline helper class to create pipelines loading modules from a flexible manifest.
"""
import argparse
from collections import defaultdict
import os
import sys
import json
import logging
import re
import webbrowser
import uuid
import shutil
import yaml
import jsonpath_ng
from typing import Callable
from hashlib import sha256
from pathlib import Path
from toposort import toposort_flatten
import datetime


try:
    import hydra
    from hydra.core.config_store import ConfigStore
    from hydra.core.hydra_config import HydraConfig
    from omegaconf import DictConfig, OmegaConf, open_dict
    from flatten_dict import flatten

    import azureml
    from azureml.core import Datastore
    from azureml.core import Experiment
    from azureml.core import Dataset
    from azureml.pipeline.core import PipelineRun
    from azure.ml.component.component import Component, Input, Output
    from azure.ml.component._core._component_definition import (
        ComponentDefinition,
        ComponentType,
    )
    from azure.ml.component.pipeline import PipelineParameter
except ImportError as error:
    raise ImportError(
        f"{error.msg}. Please install using `pip install shrike[pipeline]`."
    )

from shrike import __version__
from shrike.pipeline.aml_connect import azureml_connect, current_workspace
from shrike.pipeline.canary_helper import get_repo_info
from shrike.pipeline.module_helper import AMLModuleLoader
from shrike.pipeline.pipeline_config import default_config_dict, HDI_DEFAULT_CONF
from shrike.pipeline.telemetry_utils import TelemetryLogger
from shrike._core import is_eyesoff_helper


log = logging.getLogger(__name__)


class AMLPipelineHelper:
    """Helper class for building pipelines"""

    BUILT_PIPELINE = None  # the hydra run decorator doesn't allow for return, we're using this variable instead (hack)

    def __init__(self, config, module_loader=None):
        """Constructs the pipeline helper.

        Args:
            config (DictConfig): config for this object
            module_loader (AMLModuleLoader): which module loader to (re)use
        """
        self.config = config

        if module_loader is None:
            log.info(
                f"Creating instance of AMLModuleLoader for {self.__class__.__name__}"
            )
            self.module_loader = AMLModuleLoader(self.config)
        else:
            self.module_loader = module_loader

    ######################
    ### CUSTOM METHODS ###
    ######################

    @classmethod
    def get_config_class(cls):
        """Returns a dataclass containing config for this pipeline"""
        pass

    @classmethod
    def required_subgraphs(cls):
        """Dependencies on other subgraphs
        Returns:
            dict[str, AMLPipelineHelper]: dictionary of subgraphs used for building this one.
                keys are whatever string you want for building your graph
                values should be classes inherinting from AMLPipelineHelper.
        """
        return {}

    @classmethod
    def required_modules(cls):
        """Dependencies on modules/components

        Returns:
            dict[str, dict]: manifest
        """
        return {}

    def build(self, config):
        """Builds a pipeline function for this pipeline.

        Args:
            config (DictConfig): configuration object (see get_config_class())

        Returns:
            pipeline_function: the function to create your pipeline
        """
        raise NotImplementedError("You need to implement your build() method.")

    def pipeline_instance(self, pipeline_function, config):
        """Creates an instance of the pipeline using arguments.

        Args:
            pipeline_function (function): the pipeline function obtained from self.build()
            config (DictConfig): configuration object (see get_config_class())

        Returns:
            pipeline: the instance constructed using build() function
        """
        raise NotImplementedError(
            "You need to implement your pipeline_instance() method."
        )

    def canary(self, args, experiment, pipeline_run):
        """Tests the output of the pipeline"""
        pass

    ##################################
    ### USER FACING HELPER METHODS ###
    ##################################

    def workspace(self):
        """Gets the current workspace"""
        return current_workspace()

    def component_load(self, component_key) -> Callable[..., "Component"]:
        """Loads one component from the manifest"""
        return self.module_loader.load_module(component_key, self.required_modules())

    def module_load(self, module_key):
        """Loads one module from the manifest"""
        return self.module_loader.load_module(module_key, self.required_modules())

    def subgraph_load(self, subgraph_key, custom_config=OmegaConf.create()) -> Callable:
        """Loads one subgraph from the manifest
        Args:
            subgraph_key (str): subgraph identifier that is used in the required_subgraphs() method
            custom_config (DictConfig): custom configuration object, this custom object witll be
            added to the pipeline config

        """
        subgraph_class = self.required_subgraphs()[subgraph_key]

        subgraph_config = self.config.copy()
        if custom_config:
            with open_dict(subgraph_config):
                for key in custom_config:
                    subgraph_config[key] = custom_config[key]

        log.info(f"Building subgraph [{subgraph_key} as {subgraph_class.__name__}]...")
        # NOTE: below creates subgraph with updated pipeline_config
        subgraph_instance = subgraph_class(
            config=subgraph_config, module_loader=self.module_loader
        )
        # subgraph_instance.setup(self.pipeline_config)
        return subgraph_instance.build(subgraph_config)

    def dataset_load(self, name, version="latest"):
        """Loads a dataset by either id or name.

        Args:
            name (str): name or uuid of dataset to load
            version (str): if loading by name, used to specify version (default "latest")

        NOTE: in AzureML SDK there are 2 different methods for loading dataset
        one for id, one for name. This method just wraps them up in one."""
        # test if given name is a uuid
        try:
            parsed_uuid = uuid.UUID(name)
            log.info(f"Getting a dataset handle [id={name}]...")
            return Dataset.get_by_id(self.workspace(), id=name)
        except ValueError:
            log.info(f"Getting a dataset handle [name={name} version={version}]...")
            return Dataset.get_by_name(self.workspace(), name=name, version=version)

    @staticmethod
    def validate_experiment_name(name):
        """
        Check whether the experiment name is valid. It's required that
        experiment names must be between 1 to 250 characters, start with
        letters or numbers. Valid characters are letters, numbers, "_",
        and the "-" character.
        """
        if len(name) < 1 or len(name) > 250:
            raise ValueError("Experiment names must be between 1 to 250 characters!")
        if not re.match("^[a-zA-Z0-9]$", name[0]):
            raise ValueError("Experiment names must start with letters or numbers!")
        if not re.match("^[a-zA-Z0-9_-]*$", name):
            raise ValueError(
                "Valid experiment names must only contain letters, numbers, underscore and dash!"
            )
        return True

    @staticmethod
    def validate_connection_attribute(connection_attribute, connection_attribute_type):
        """
        Function that will verify whether the given connection_attribute (Workspace name, RG name, or Subscription Id) matches the constraints on length, allowed characters, etc...
        """

        # initialize min_length and max_length
        min_length = 0
        max_length = sys.maxsize

        # populate the constraints based on connection_attribute_type
        if connection_attribute_type == "workspace":
            min_length = 3
            max_length = 33
            # from https://docs.microsoft.com/en-us/azure/templates/microsoft.machinelearningservices/workspaces?tabs=bicep#workspaces
            connection_attribute_type_constraints = (
                "only alphanumeric characters and hyphens"
            )
            regex_target = "^[a-zA-Z0-9-]*$"
        elif connection_attribute_type == "resource_group":
            # from https://docs.microsoft.com/en-us/azure/azure-resource-manager/management/resource-name-rules#microsoftresources
            connection_attribute_type_constraints = "alphanumeric, underscore, parentheses, hyphen, period (except at end), and Unicode characters that match the allowed characters"
            regex_target = "^[-\w\._\(\)]+[-\w_\(\)]$"
        elif connection_attribute_type == "subscription_id":
            # can't find an official link for this...
            connection_attribute_type_constraints = "V4 GUID"
            # following regex taken from: https://stackoverflow.com/questions/19989481/how-to-determine-if-a-string-is-a-valid-v4-uuid
            regex_target = "^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$"
        else:
            regex_target = ""
            connection_attribute_type_constraints = ""
            raise ValueError(
                "connection_attribute_type must be one of 'workspace', 'resource_group', or 'subscription_id'"
            )

        # check the length first
        if (len(connection_attribute) < min_length) or (
            len(connection_attribute) > max_length
        ):
            raise ValueError(
                f"{connection_attribute} must be between {min_length} and {max_length} characters."
            )

        # check whether the regex matches
        if not re.match(regex_target, connection_attribute):
            raise ValueError(
                f"Valid {connection_attribute_type} names have the following constraints: {connection_attribute_type_constraints}"
            )

        return True

    def is_eyesoff(self) -> bool:
        """ "
        Check whether the workspace is eyes-off.
        If it lives in a non-Torus tenant, then eyes-off;
        If in Torus tenant, check whether it is in the allow-list of eyes-on Torus subscriptions.
        """
        workspace = self.workspace()
        tenant_id = workspace.get_details()["identity"]["tenant_id"]
        subscription_id = workspace.subscription_id
        return is_eyesoff_helper(tenant_id, subscription_id)

    #######################
    ### HELPER BACKEND  ###
    #######################

    @classmethod
    def _default_config(cls):
        """Builds the default config for the pipeline class"""
        config_store = ConfigStore.instance()

        config_dict = default_config_dict()
        cls._build_config(config_dict)

        config_store.store(name="default", node=config_dict)
        return OmegaConf.structured(config_dict)

    @classmethod
    def _build_config(cls, config_dict, modules_config=None):
        """Builds the entire configuration object for this graph and all subgraphs."""
        self_config_class = cls.get_config_class()
        if self_config_class:
            config_dict[self_config_class.__name__] = self_config_class

        for subgraph_key, subgraph_class in cls.required_subgraphs().items():
            subgraph_class._build_config(config_dict)

    def _set_all_inputs_to(self, module_instance, input_mode):
        """Sets all module inputs to a given intput mode"""
        input_names = [
            a
            for a in dir(module_instance.inputs)
            if isinstance(getattr(module_instance.inputs, a), Input)
        ]
        for input_key in input_names:
            input_instance = getattr(module_instance.inputs, input_key)
            input_instance.configure(mode=input_mode)
            log.info(f"Configured input {input_key} to use mode {input_mode}")

    def _set_all_outputs_to(
        self, module_instance, output_mode, compliant=True, datastore_name=None
    ):
        """Sets all module outputs to a given output mode"""
        output_names = [
            a
            for a in dir(module_instance.outputs)
            if isinstance(getattr(module_instance.outputs, a), Output)
        ]
        if not datastore_name:
            datastore_name = (
                self.config.compute.compliant_datastore
                if compliant
                else self.config.compute.noncompliant_datastore
            )
        for output_key in output_names:
            output_instance = getattr(module_instance.outputs, output_key)
            if output_mode is None:
                output_instance.configure(datastore=datastore_name)
            else:
                output_instance.configure(
                    datastore=datastore_name,
                    output_mode=output_mode,
                )
            log.info(
                f"Configured output {output_key} to use mode {output_mode} and datastore {datastore_name}"
            )

    def _apply_windows_runsettings(
        self,
        module_name,
        module_instance,
        mpi=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        input_mode=None,
        output_mode=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a windows module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            mpi (bool): is job mpi ?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): force process_count_per_node over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.module_loader.is_local(module_name):
            target = (
                target
                if target is not None
                else self.config.compute.windows_cpu_dc_target
            )
        else:
            target = (
                target
                if target is not None
                else self.config.compute.windows_cpu_prod_target
            )

        log.info(
            f"Using windows compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(target=target, **custom_runtime_arguments)
        if mpi:
            node_count = node_count if node_count is not None else 1
            process_count_per_node = (
                process_count_per_node if process_count_per_node is not None else 1
            )
            log.info(
                f"Using mpi with node_count={node_count} process_count_per_node={process_count_per_node}"
            )
            module_instance.runsettings.resource_layout.configure(
                node_count=node_count,
                process_count_per_node=process_count_per_node,
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        else:
            self._set_all_inputs_to(
                module_instance, self.config.compute.windows_input_mode
            )

        if output_mode:
            self._set_all_outputs_to(module_instance, output_mode)
        else:
            self._set_all_outputs_to(
                module_instance, self.config.compute.windows_output_mode
            )

    def _apply_hdi_runsettings(
        self,
        module_name,
        module_instance,
        target=None,
        driver_cores=None,
        driver_memory=None,
        executor_memory=None,
        executor_cores=None,
        number_executors=None,
        conf=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a HDI module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            target (str): force target compute over hydra conf
            driver_cores (int): force driver_cores over hydra conf
            driver_memory (str): force driver_memory over hydra conf
            executor_memory (int): force executor_memory over hydra conf
            executor_cores (int): force executor_cores over hydra conf
            number_executors (int): force number_executors over hydra conf
            conf (str): force conf over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.module_loader.is_local(module_name):
            target = target if target is not None else self.config.compute.hdi_dc_target
            if not self.config.compute.hdi_dc_target:
                raise Exception(
                    f"Your HDI component {module_name} is using local version. Please specify hdi_dc_target"
                )
        else:
            target = (
                target if target is not None else self.config.compute.hdi_prod_target
            )
        log.info(
            f"Using HDI compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        merged_conf = json.loads(HDI_DEFAULT_CONF)
        new_conf = (
            self.config.compute.hdi_conf if "hdi_conf" in self.config.compute else None
        )
        if conf is not None:
            new_conf = conf
        if new_conf is not None:
            if isinstance(new_conf, str):
                new_conf = json.loads(new_conf)
            elif isinstance(new_conf, DictConfig):
                new_conf = flatten(dict(new_conf), reducer="dot")
            else:
                raise ValueError(
                    "computed.hdi_conf is not a valid json string or a single tested configuration."
                )
            merged_conf.update(new_conf)

        module_instance.runsettings.configure(target=target)

        module_instance.runsettings.hdinsight.configure(
            driver_memory=driver_memory
            if driver_memory is not None
            else self.config.compute.hdi_driver_memory,
            driver_cores=driver_cores
            if driver_cores is not None
            else self.config.compute.hdi_driver_cores,
            executor_memory=executor_memory
            if executor_memory is not None
            else self.config.compute.hdi_executor_memory,
            executor_cores=executor_cores
            if executor_cores is not None
            else self.config.compute.hdi_executor_cores,
            number_executors=number_executors
            if number_executors is not None
            else self.config.compute.hdi_number_executors,
            conf=merged_conf,
            **custom_runtime_arguments,
        )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, datastore_name=datastore_name
        )

    def _apply_parallel_runsettings(
        self,
        module_name,
        module_instance,
        windows=False,
        gpu=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        mini_batch_size=None,
        run_invocation_timeout=None,
        run_max_try=None,
        error_threshold=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a ParallelRunStep linux module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            windows (bool): is the module using Windows compute?
            gpu (bool): is the module using GPU compute?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): force process_count_per_node over hydra conf
            mini_batch_size (int): force mini_batch_size over hydra conf
            run_invocation_timeout (int): force run_invocation_timeout over hydra conf
            run_max_try (int): force run_max_try over hydra conf
            error_threshold (int): The number of file failures for the input FileDataset that should be ignored during processing.
                If the error count goes above this value, then the job will be aborted.
                Error threshold is for the entire input and not for individual mini-batches sent to run() method.
                The range is [-1, int.max]. -1 indicates ignoring all failures during processing.
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.module_loader.is_local(module_name):
            if windows:
                if gpu:
                    raise ValueError(
                        "A GPU compute target with Windows OS is not available yet!"
                    )
                else:
                    _target = self.config.compute.windows_cpu_dc_target
            else:
                if gpu:
                    _target = self.config.compute.linux_gpu_dc_target
                else:
                    _target = self.config.compute.linux_cpu_dc_target
        else:
            if windows:
                if gpu:
                    raise ValueError(
                        "A GPU compute target with Windows OS is not available yet!"
                    )
                else:
                    _target = self.config.compute.windows_cpu_prod_target
            else:
                if gpu:
                    _target = self.config.compute.linux_gpu_prod_target
                else:
                    _target = self.config.compute.linux_cpu_prod_target

        target = target if target is not None else _target

        log.info(
            f"Using parallelrunstep compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(target=target)

        module_instance.runsettings.parallel.configure(
            node_count=node_count
            if node_count is not None
            else self.config.compute.parallel_node_count,
            process_count_per_node=process_count_per_node
            if process_count_per_node is not None
            else (
                self.config.compute.parallel_process_count_per_node
                if "parallel_process_count_per_node" in self.config.compute
                else None
            ),
            mini_batch_size=str(
                mini_batch_size
                if mini_batch_size is not None
                else self.config.compute.parallel_mini_batch_size
            ),
            run_invocation_timeout=run_invocation_timeout
            if run_invocation_timeout is not None
            else self.config.compute.parallel_run_invocation_timeout,
            run_max_try=run_max_try
            if run_max_try is not None
            else self.config.compute.parallel_run_max_try,
            error_threshold=error_threshold
            if error_threshold is not None
            else self.config.compute.parallel_error_threshold,
            **custom_runtime_arguments,
        )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        if output_mode:
            self._set_all_outputs_to(
                module_instance, output_mode, datastore_name=datastore_name
            )

    def _apply_linux_runsettings(
        self,
        module_name,
        module_instance,
        mpi=False,
        gpu=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for linux module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            mpi (bool): is the job mpi?
            gpu (bool): is the job using GPU?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): force process_count_per_node over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.module_loader.is_local(module_name) and gpu:
            target = (
                target
                if target is not None
                else self.config.compute.linux_gpu_dc_target
            )
            log.info(
                f"Using target {target} for local code GPU module {module_name} from pipeline class {self.__class__.__name__}"
            )
        elif not self.module_loader.is_local(module_name) and gpu:
            target = (
                target
                if target is not None
                else self.config.compute.linux_gpu_prod_target
            )
            log.info(
                f"Using target {target} for registered GPU module {module_name} from pipeline class {self.__class__.__name__}"
            )
        elif self.module_loader.is_local(module_name) and not gpu:
            target = (
                target
                if target is not None
                else self.config.compute.linux_cpu_dc_target
            )
            log.info(
                f"Using target {target} for local CPU module {module_name} from pipeline class {self.__class__.__name__}"
            )
        elif not self.module_loader.is_local(module_name) and not gpu:
            target = (
                target
                if target is not None
                else self.config.compute.linux_cpu_prod_target
            )
            log.info(
                f"Using target {target} for registered CPU module {module_name} from pipeline class {self.__class__.__name__}"
            )

        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(target=target, **custom_runtime_arguments)

        if mpi:
            node_count = node_count if node_count is not None else 1
            process_count_per_node = (
                process_count_per_node if process_count_per_node is not None else 1
            )
            log.info(
                f"Using mpi with node_count={node_count} process_count_per_node={process_count_per_node}"
            )
            module_instance.runsettings.resource_layout.configure(
                node_count=node_count,
                process_count_per_node=process_count_per_node,
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        else:
            self._set_all_inputs_to(
                module_instance, self.config.compute.linux_input_mode
            )
        if output_mode:
            self._set_all_outputs_to(
                module_instance, output_mode, datastore_name=datastore_name
            )
        else:
            self._set_all_outputs_to(
                module_instance,
                self.config.compute.linux_output_mode,
                datastore_name=datastore_name,
            )

    def _apply_scope_runsettings(
        self,
        module_name,
        module_instance,
        input_mode=None,
        output_mode=None,
        scope_param=None,
        custom_job_name_suffix=None,
        adla_account_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a scope module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            scope_param (str): Parameters to pass to scope e.g. Nebula parameters, VC allocation parameters etc.
            custom_job_name_suffix (str): Optional parameter defining custom string to append to job name
            adla_account_name (str): The name of the Cosmos-migrated Azure Data Lake Analytics account to submit scope job
            custom_runtime_arguments (dict): any additional custom args
        """
        log.info(
            f"Using scope compute target to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(module_instance, output_mode, compliant=False)

        module_instance.runsettings.scope.configure(
            adla_account_name=adla_account_name,
            scope_param=scope_param,
            custom_job_name_suffix=custom_job_name_suffix,
        )

    def _apply_datatransfer_runsettings(
        self,
        module_name,
        module_instance,
        compliant=True,
        target=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a HDI module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            compliant (bool): destination datastore, `compliant_datastore` if True, else `noncompliant_datastore`
            target (str): force target compute over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """
        log.info(
            f"Using datatransfer compute target {self.config.compute.datatransfer_target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(
            target=target
            if target is not None
            else self.config.compute.datatransfer_target,
        )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, compliant, datastore_name
        )

    def _apply_sweep_runsettings(
        self,
        module_name,
        module_instance,
        windows=False,
        gpu=False,
        target=None,
        input_mode=None,
        output_mode=None,
        node_count=None,
        process_count_per_node=None,
        algorithm=None,
        primary_metric=None,
        goal=None,
        policy_type=None,
        evaluation_interval=None,
        delay_evaluation=None,
        slack_factor=None,
        slack_amount=None,
        truncation_percentage=None,
        max_total_trials=None,
        max_concurrent_trials=None,
        timeout_minutes=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a sweep component.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            target (str): force target compute over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf

            # For below sweep specific parameters configurations, see below doc link for more info:
            # https://componentsdk.azurewebsites.net/components/sweep_component.html#set-runsettings
            algorithm (str): sweep sampling method
            primary_metric (str): the primary metric of the hyperparameter tuning to optimize
            goal (str): whether the primary metric will be maximize or minimize when evaluating the trials
            policy_type (str): sweep early termination policy type
            evaluation_interval (int): frequency of applying the policy
            delay_evaluation (int): delays the first policy evaluation for a specified number of intervals
            slack_factor (float): the slack allowed with respect to the best performing training run, as a ratio
            slack_amount (float): the slack allowed with respect to the best performing training run, as an absolute ampunt. You should only specify either slack_factor or slack_amount, but not both.
            truncation_percentage (int): the percentage of lowest performing runs to terminate at each evaluation interval. An integer value between 1 and 99.
            max_total_trials (int): maximum number of trial runs. Must be an integer between 1 and 1000.
            max_concurrent_trials (int): maximum number of runs that can run concurrently. If not specified, all runs launch in parallel. If specified, must be an integer between 1 and 100.
            timeout_minutes (int): maximum duration, in minutes, of the hyperparameter tuning experiment. Runs after this duration are canceled.

            custom_runtime_arguments (dict): any additional custom args
        """
        if target is not None:
            target = target
        elif self.module_loader.is_local(module_name):
            if windows:
                if gpu:
                    raise ValueError(
                        "A GPU compute target with Windows OS is not available yet!"
                    )
                else:
                    target = self.config.compute.windows_cpu_dc_target
            else:
                if gpu:
                    target = self.config.compute.linux_gpu_dc_target
                else:
                    target = self.config.compute.linux_cpu_dc_target
        else:
            if windows:
                if gpu:
                    raise ValueError(
                        "A GPU compute target with Windows OS is not available yet!"
                    )
                else:
                    target = self.config.compute.windows_cpu_prod_target
            else:
                if gpu:
                    target = self.config.compute.linux_gpu_prod_target
                else:
                    target = self.config.compute.linux_cpu_prod_target
        log.info(
            f"Using target {target} to run sweep component {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            log.info(f"Adding custom runtime arguments {custom_runtime_arguments}")

        module_instance.runsettings.configure(target=target, **custom_runtime_arguments)

        if node_count:
            log.info(
                f"Setting node_count={node_count} and process_count_per_node={process_count_per_node} as run settings for sweep component {module_name} from pipeline class {self.__class__.__name__}"
            )
            module_instance.runsettings.resource_layout.configure(
                node_count=node_count, process_count_per_node=process_count_per_node
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        if output_mode:
            self._set_all_outputs_to(
                module_instance, output_mode, datastore_name=datastore_name
            )

        if algorithm:
            module_instance.runsettings.sweep.algorithm = algorithm
        if primary_metric:
            module_instance.runsettings.sweep.objective.configure(
                primary_metric=primary_metric
            )
        if goal:
            module_instance.runsettings.sweep.objective.configure(goal=goal)
        if policy_type:
            module_instance.runsettings.sweep.early_termination.configure(
                policy_type=policy_type
            )
        if evaluation_interval:
            module_instance.runsettings.sweep.early_termination.configure(
                evaluation_interval=evaluation_interval
            )
        if delay_evaluation:
            module_instance.runsettings.sweep.early_termination.configure(
                delay_evaluation=delay_evaluation
            )
        if slack_factor:
            module_instance.runsettings.sweep.early_termination.configure(
                slack_factor=slack_factor
            )
        if slack_amount:
            module_instance.runsettings.sweep.early_termination.configure(
                slack_amount=slack_amount
            )
        if truncation_percentage:
            module_instance.runsettings.sweep.early_termination.configure(
                truncation_percentage=truncation_percentage
            )
        if max_total_trials:
            module_instance.runsettings.sweep.limits.configure(
                max_total_trials=max_total_trials
            )
        if max_concurrent_trials:
            module_instance.runsettings.sweep.limits.configure(
                max_concurrent_trials=max_concurrent_trials
            )
        if timeout_minutes:
            module_instance.runsettings.sweep.limits.configure(
                timeout_minutes=timeout_minutes
            )

    def _check_module_runsettings_consistency(self, module_key, module_instance):
        """Verifies if entry at module_key matches the module instance description"""
        (
            module_manifest_entry,
            _,
            is_in_manifest,
        ) = self.module_loader.get_module_manifest_entry(
            module_key, modules_manifest=self.required_modules()
        )

        if "name" in module_manifest_entry:
            if "source" in module_manifest_entry:
                log.warning(
                    "!!IMPORTANT: Key word `source` is deprecated for module artifacts. Please remove it."
                )
            if module_manifest_entry["name"] == module_instance.name:
                return
            if "namespace" in module_manifest_entry:
                module_entry_name = (
                    module_manifest_entry["namespace"]
                    + "://"
                    + module_manifest_entry["name"]
                )
                if module_entry_name == module_instance.name:
                    return
            raise Exception(
                f"During build() of graph class {self.__class__.__name__}, call to self.apply_recommended_runsettings() is wrong: key used as first argument ('{module_key}') maps to a module reference {module_manifest_entry} which name is different from the module instance provided as 2nd argument (name={module_instance.name}), did you use the wrong module key as first argument?"
            )

    def _get_component_name_from_instance(self, component_instance):
        component_manifest_list = self.config.modules.manifest
        component_name = component_instance.name
        for component_manifest_entry in component_manifest_list:
            try:
                if component_manifest_entry["name"] == component_name:
                    return component_manifest_entry["key"] or component_name
                if "namespace" in component_manifest_entry:
                    component_entry_name = (
                        component_manifest_entry["namespace"]
                        + "://"
                        + component_manifest_entry["name"]
                    )
                    if component_entry_name == component_name:
                        return component_manifest_entry["key"] or component_name
            except:
                pass
        raise ValueError(
            f"Could not find component matching {component_name}. Please check your spelling."
        )

    def apply_smart_runsettings(
        self,
        component_instance,
        gpu=False,  # can't autodetect that
        hdi="auto",
        windows="auto",
        parallel="auto",
        mpi="auto",
        scope="auto",
        datatransfer="auto",
        sweep="auto",
        **custom_runtime_arguments,
    ):
        """Applies regular settings for a given component.

        Args:
            component_instance (Component): the AML component we need to add settings to
            gpu (bool): is the component using GPU?
            hdi (bool): is the component using HDI/Spark?
            windows (bool): is the component using Windows compute?
            parallel (bool): is the component using ParallelRunStep?
            mpi (bool): is the component using Mpi?
            scope (bool): is the component using scope?
            datatransfer (bool): is the component using datatransfer?
            sweep (bool): is the component using sweep?
            custom_runtime_arguments (dict): any additional custom args
        """
        # infer component_name
        component_name = self._get_component_name_from_instance(component_instance)
        self.apply_recommended_runsettings(
            component_name,
            component_instance,
            gpu,
            hdi,
            windows,
            parallel,
            mpi,
            scope,
            datatransfer,
            sweep,
            **custom_runtime_arguments,
        )

    def apply_recommended_runsettings(
        self,
        module_name,
        module_instance,
        gpu=False,  # can't autodetect that
        hdi="auto",
        windows="auto",
        parallel="auto",
        mpi="auto",
        scope="auto",
        datatransfer="auto",
        sweep="auto",
        **custom_runtime_arguments,
    ):
        """Applies regular settings for a given module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            gpu (bool): is the module using GPU?
            hdi (bool): is the module using HDI/Spark?
            windows (bool): is the module using Windows compute?
            parallel (bool): is the module using ParallelRunStep?
            mpi (bool): is the module using Mpi?
            custom_runtime_arguments (dict): any additional custom args
        """
        # verifies if module_name corresponds to module_instance
        self._check_module_runsettings_consistency(module_name, module_instance)

        # Auto detect runsettings
        if hdi == "auto":
            hdi = str(module_instance.type) == "HDInsightComponent"
            if hdi:
                log.info(f"Module {module_name} detected as HDI: {hdi}")

        if parallel == "auto":
            parallel = str(module_instance.type) == "ParallelComponent"
            if parallel:
                log.info(f"Module {module_name} detected as PARALLEL: {parallel}")

        if mpi == "auto":
            mpi = str(module_instance.type) == "DistributedComponent"
            if mpi:
                log.info(f"Module {module_name} detected as MPI: {mpi}")

        if scope == "auto":
            scope = str(module_instance.type) == "ScopeComponent"
            if scope:
                log.info(f"Module {module_name} detected as SCOPE: {scope}")

        if sweep == "auto":
            sweep = str(module_instance.type) == "SweepComponent"
            if sweep:
                log.info(f"Module {module_name} detected as SweepComponent: {sweep}")

        if windows == "auto":
            if (
                str(module_instance.type) == "HDInsightComponent"
                or str(module_instance.type) == "ScopeComponent"
                or str(module_instance.type) == "DataTransferComponent"
                or str(module_instance.type) == "SweepComponent"
            ):
                # HDI/scope/datatransfer/sweep modules might not have that environment object
                windows = False
            else:
                windows = (
                    module_instance._definition.environment.os.lower() == "windows"
                )
                if windows:
                    log.info(f"Module {module_name} detected as WINDOWS: {windows}")

        if datatransfer == "auto":
            datatransfer = str(module_instance.type) == "DataTransferComponent"
            if datatransfer:
                log.info(
                    f"Module {module_name} detected as DATATRANSFER: {datatransfer}"
                )

        if parallel:
            self._apply_parallel_runsettings(
                module_name,
                module_instance,
                windows=windows,
                gpu=gpu,
                **custom_runtime_arguments,
            )
            return

        if sweep:
            self._apply_sweep_runsettings(
                module_name,
                module_instance,
                windows=windows,
                gpu=gpu,
                **custom_runtime_arguments,
            )
            return

        if windows:
            self._apply_windows_runsettings(
                module_name,
                module_instance,
                mpi=mpi,
                **custom_runtime_arguments,
            )
            return

        if hdi:
            self._apply_hdi_runsettings(
                module_name,
                module_instance,
                **custom_runtime_arguments,
            )
            return

        if scope:
            self._apply_scope_runsettings(
                module_name, module_instance, **custom_runtime_arguments
            )
            return

        if datatransfer:
            self._apply_datatransfer_runsettings(
                module_name, module_instance, **custom_runtime_arguments
            )
            return

        self._apply_linux_runsettings(
            module_name,
            module_instance,
            mpi=mpi,
            gpu=gpu,
            **custom_runtime_arguments,
        )

    def _parse_pipeline_tags(self):
        """Parse the tags specified in the pipeline yaml"""
        pipeline_tags = {}
        if self.config.run.tags:
            if isinstance(self.config.run.tags, str):
                try:
                    # json.load the tags string in the config
                    pipeline_tags = json.loads(self.config.run.tags)
                except ValueError:
                    log.warning(
                        f"The pipeline tags {self.config.run.tags} is not a valid json-style string."
                    )
            elif isinstance(self.config.run.tags, DictConfig):
                # NOTE: values for tags need to be str
                for key, value in self.config.run.tags.items():
                    pipeline_tags[key] = str(value)
            else:
                log.warning(
                    f"The pipeline tags {self.config.run.tags} is not a valid DictConfig or json-style string."
                )
        return pipeline_tags

    def _check_if_spec_yaml_override_is_needed(self):
        if self.config.module_loader.use_local == "":
            log.info(
                "All components are using remote copy, so override will not be executed. For components you want submission-time override of images/tags/etc., please specify them in `use_local`."
            )
            return False, None
        if not self.config.tenant_overrides.allow_override:
            log.info(
                "Spec yaml file override is not allowed. If you want to use this feature, please set `tenant_overrides.allow_override` to True in your pipeline yaml."
            )
            return False, None
        cur_tenant = self.config.aml.tenant
        for tenant in self.config.tenant_overrides.mapping.keys():
            if tenant == cur_tenant:
                log.info(
                    f"Your tenant is inconsistent with spec yaml, We will override relevant fields in your spec yaml files based on entry `{tenant}` in your pipeline yaml file."
                )
                return True, self.config.tenant_overrides.mapping[tenant]
            if self.config.run.config_dir:
                config_file_path = os.path.join(
                    self.config.run.config_dir, "aml", tenant + ".yaml"
                )
                log.info(f"config_file_path is {config_file_path}")
                if os.path.exists(config_file_path):
                    with open(config_file_path, "r") as file:
                        config = yaml.safe_load(file)
                    if config["tenant"] == cur_tenant:
                        log.info(
                            f"Your tenant is inconsistent with spec yaml, We will override relevant fields in your spec yaml files based on entry `{config_file_path}` in your pipeline yaml file."
                        )
                        return True, self.config.tenant_overrides.mapping[tenant]
        return False, None

    def _override_spec_yaml(self, spec_mapping):
        module_keys = self.module_loader.modules_manifest
        yaml_to_be_recovered = []
        env_yaml_override_is_needed = (
            spec_mapping.remove_polymer_pkg_idx
            if "remove_polymer_pkg_idx" in spec_mapping
            else False
        )
        for module_key in module_keys:
            if not self.module_loader.is_local(module_key):
                log.info(
                    f"Component {module_key} is using the remote copy. Skipping overrides."
                )
                continue
            log.info(f"Overriding for component: {module_key}.")
            module_entry = self.module_loader.modules_manifest[module_key]
            spec_path = os.path.join(
                self.module_loader.local_steps_folder, module_entry["yaml"]
            )
            (
                old_spec_path,
                old_env_file_path,
                new_env_file_path,
            ) = self._override_single_spec_yaml(
                spec_path, spec_mapping, env_yaml_override_is_needed
            )

            yaml_to_be_recovered.append(
                [old_spec_path, spec_path, old_env_file_path, new_env_file_path]
            )
        return yaml_to_be_recovered

    def _update_value_given_flattened_key(self, nested_dict, dot_key, new_val):
        log.info(f"Updating key {dot_key}")
        split_key = dot_key.split(".")
        res = nested_dict
        path = ""
        for key in split_key[:-1]:
            path += key + "."
            if not isinstance(res, dict) or key not in res:
                raise KeyError(
                    f"Key {dot_key} not in {nested_dict}. It failed at {path}."
                )
            res = res[key]
        if isinstance(res, dict) and split_key[-1] in res:
            res[split_key[-1]] = new_val
            log.info(f"The field {dot_key} has been updated to {new_val} successfully.")
        else:
            raise KeyError(f"Key {dot_key} not in {nested_dict}.")

    def _override_single_spec_yaml(
        self, spec_path, spec_mapping, env_yaml_override_is_needed
    ):
        spec_filename, spec_ext = os.path.splitext(spec_path)
        old_spec_path = (
            spec_filename + ".not_used"
        )  # remove ".yaml" extension to avoid confusion
        shutil.copyfile(
            spec_path, old_spec_path
        )  # need to recover after pipeline submission

        with open(spec_path) as file:
            spec = yaml.safe_load(file)
        for key in spec_mapping:
            match = jsonpath_ng.parse(key).find(spec)
            if match:
                try:
                    log.info(f"Find a matching field to override: {key}.")
                    original_val = match[0].value
                    log.info(
                        f"Original value is {original_val}. Looking for new value now..."
                    )
                    spec_mapping_to_use = spec_mapping[key]
                    if isinstance(original_val, str):
                        log.info(f"The field to be updated is str.")
                        if original_val in spec_mapping_to_use:
                            new_val = spec_mapping_to_use[original_val]
                            log.info(f"The new value is: {new_val}.")
                            self._update_value_given_flattened_key(spec, key, new_val)
                        else:
                            for pattern in spec_mapping_to_use:
                                if re.match(pattern, original_val):
                                    new_val = spec_mapping_to_use[pattern]
                                    log.info(f"The new pattern is: {new_val}.")
                                    self._update_value_given_flattened_key(
                                        spec,
                                        key,
                                        re.sub(pattern, new_val, original_val),
                                    )
                    elif isinstance(original_val, dict):
                        log.info("The field to be updated is dict")
                        for spec_mapping_key in spec_mapping_to_use:
                            self._update_value_given_flattened_key(
                                spec,
                                ".".join([key, spec_mapping_key]),
                                spec_mapping_to_use[spec_mapping_key],
                            )
                    else:
                        log.info(
                            f"Override for key {key} is not supported yet. Please open a [feature request](https://github.com/Azure/shrike/issues) if necessary."
                        )
                except KeyError:
                    log.info(f"Key {key} does not in file {spec_path}. Skip overrides.")
        new_env_file_path = None
        old_env_file_path = None
        if env_yaml_override_is_needed:
            spec_dirname = os.path.dirname(spec_path)
            polymer_pkg_idx = "--index-url https://o365exchange.pkgs.visualstudio.com/_packaging/PolymerPythonPackages/pypi/simple/"
            log.info(f"Will remove {polymer_pkg_idx} from environment.conda.")
            try:
                conda_dependencies_file = spec["environment"]["conda"][
                    "conda_dependencies_file"
                ]
                log.info("conda_dependencies_file exists.")
                (
                    found_index_url,
                    new_file,
                    new_env_file_path,
                    old_env_file_path,
                ) = self._remove_polymer_pkg_idx_if_exists_and_save_new(
                    spec_dirname, conda_dependencies_file, polymer_pkg_idx
                )
                if found_index_url:
                    spec["environment"]["conda"]["conda_dependencies_file"] = new_file
            except KeyError:
                # conda_dependencies_file does not exist
                pass
            try:
                conda_dependencies = spec["environment"]["conda"]["conda_dependencies"][
                    "dependencies"
                ]
                log.info("conda_dependencies_file exists.")
                for idx, dependency in enumerate(conda_dependencies):
                    if isinstance(dependency, dict) and "pip" in dependency:
                        pip_dependencies = dependency["pip"]
                        if polymer_pkg_idx in pip_dependencies:
                            pip_dependencies.remove(polymer_pkg_idx)
                            dependency["pip"] = pip_dependencies
                            conda_dependencies[idx] = dependency
                spec["environment"]["conda"]["conda_dependencies"][
                    "dependencies"
                ] = conda_dependencies
            except KeyError:
                # conda_dependencies does not exist
                pass
            try:
                pip_requirements_file = spec["environment"]["conda"][
                    "pip_requirements_file"
                ]
                log.info("pip_requirements_file exists.")
                (
                    found_index_url,
                    new_file,
                    new_env_file_path,
                    old_env_file_path,
                ) = self._remove_polymer_pkg_idx_if_exists_and_save_new(
                    spec_dirname, pip_requirements_file, polymer_pkg_idx
                )
                if found_index_url:
                    spec["environment"]["conda"]["pip_requirements_file"] = new_file
            except KeyError:
                # pip_requirements_file does not exist
                pass

        with open(spec_path, "w") as file:
            yaml.safe_dump(spec, file)
        return old_spec_path, old_env_file_path, new_env_file_path

    def _remove_polymer_pkg_idx_if_exists_and_save_new(
        self, spec_dirname, file, polymer_pkg_idx
    ):
        found_index_url = False
        new_file = ""
        new_file_path = ""
        old_file_path = ""
        with open(os.path.join(spec_dirname, file), "r") as f:
            lines = f.readlines()
        for line in lines:
            if polymer_pkg_idx in line:
                found_index_url = True
                lines.remove(line)
        if found_index_url:
            file_name, file_ext = os.path.splitext(file)
            new_file = file_name + "_" + self.config.aml.tenant + file_ext
            new_file_path = os.path.join(spec_dirname, new_file)
            with open(new_file_path, "w") as f:
                f.writelines(lines)
            old_file_path = os.path.join(
                spec_dirname, os.path.splitext(file)[0] + ".not_used"
            )
            shutil.move(os.path.join(spec_dirname, file), old_file_path)
        return found_index_url, new_file, new_file_path, old_file_path

    def _recover_spec_yaml(self, spec_file_pairs, keep_modified_files):
        log.info(
            f"Reverting changes to spec yaml files. Keeping modified spec yaml files: {keep_modified_files}."
        )
        for (
            old_spec_path,
            new_spec_path,
            old_env_file_path,
            new_env_file_path,
        ) in spec_file_pairs:
            # Question: do we want to keep a copy of the modified files?
            if keep_modified_files:
                filename, ext = os.path.splitext(new_spec_path)
                shutil.move(
                    new_spec_path, filename + "_" + self.config.aml.tenant + ext
                )
                if os.path.exists(filename + ".additional_includes"):
                    shutil.copyfile(
                        filename + ".additional_includes",
                        filename
                        + "_"
                        + self.config.aml.tenant
                        + ".additional_includes",
                    )
            else:
                if new_env_file_path:
                    os.remove(new_env_file_path)
            shutil.move(old_spec_path, new_spec_path)
            if old_env_file_path:
                shutil.move(
                    old_env_file_path, os.path.splitext(old_env_file_path)[0] + ".yaml"
                )

    def _recover_tenant_overrides(
        self, override, yaml_to_be_recovered, keep_modified_files
    ):
        if override and yaml_to_be_recovered:
            try:
                self._recover_spec_yaml(yaml_to_be_recovered, keep_modified_files)
            except BaseException as e:
                log.error(f"An error occurred, recovery is not successful: {e}")

    def _update_hdi_dc_configs(self, pipeline):
        if not self.config.compute.hdi_dc_target:
            log.info("This pipeline doesn't involve HDI DC.")
            return pipeline
        nodes = pipeline._expand_pipeline_nodes()
        graph = defaultdict(list)
        node_mapping = {}
        for n in nodes:
            node_mapping[n._instance_id] = n
            for input in n.inputs.values():
                try:
                    owner = input._get_data_owner()
                    if owner:
                        graph[n._instance_id].append(owner._instance_id)
                except AttributeError:
                    continue
        sorted_node_ids = toposort_flatten(graph)
        use_hdi_dc = defaultdict(bool)
        for id in sorted_node_ids:
            if id not in node_mapping:
                continue
            node = node_mapping[id]
            if (
                node.type != "ScopeComponent"
                and node.runsettings.target == self.config.compute.hdi_dc_target
            ):
                use_hdi_dc[id] = True
                continue
            for parent in graph[id]:
                if use_hdi_dc[parent]:
                    use_hdi_dc[id] = True
                    break

        for node in nodes:
            if use_hdi_dc[node._instance_id]:
                if node.type == "ScopeComponent":
                    # scope component doesn't need to configure compute target
                    # output datastore has to be cosmos (ADLSg1), while dc_datastore is ADLSg2
                    continue
                self._set_all_outputs_to(
                    node,
                    output_mode=None,
                    datastore_name=self.config.compute.dc_datastore,
                )
                log.info(
                    f"Writing outputs of {node.name} to dc_datastore {self.config.compute.dc_datastore}"
                )
                if node.runsettings.target == self.config.compute.hdi_prod_target:
                    node.runsettings.configure(target=self.config.compute.hdi_dc_target)
                elif (
                    node.runsettings.target == self.config.compute.linux_cpu_prod_target
                ):
                    node.runsettings.configure(
                        target=self.config.compute.linux_cpu_dc_target
                    )
                elif (
                    node.runsettings.target == self.config.compute.linux_gpu_prod_target
                ):
                    node.runsettings.configure(
                        target=self.config.compute.linux_gpu_dc_target
                    )
                elif (
                    node.runsettings.target
                    == self.config.compute.windows_cpu_prod_target
                ):
                    node.runsettings.configure(
                        target=self.config.compute.windows_cpu_dc_target
                    )
                else:
                    continue
                log.info(
                    f"Updating compute target for {node.name} {node._instance_id} to {node.runsettings.target}"
                )
        return pipeline

    def _publish_to_endpoint(self, pipeline):
        name = datetime.datetime.now()
        endpoint_name = self.config.run.endpoint_name or self.config.run.experiment_name
        endpoint_description = (
            self.config.run.endpoint_description
            or self.config.run.experiment_description
        )

        pipeline._publish_to_endpoint(
            experiment_name=self.config.run.experiment_name,
            name=name,
            pipeline_endpoint_name=endpoint_name,
            description=endpoint_description,
            workspace=self.workspace(),
        )
        log.info(
            f"Publishing pipeline as endpoint {endpoint_name} with name {name} to workspace {self.workspace()}."
        )

    ################
    ### MAIN/RUN ###
    ################

    def connect(self):
        """Connect to the AML workspace using internal config"""
        # Only call azureml_connect if there is not an already a pre-existing workspace
        try:
            return current_workspace()
        except:
            return azureml_connect(
                aml_subscription_id=self.config.aml.subscription_id,
                aml_resource_group=self.config.aml.resource_group,
                aml_workspace_name=self.config.aml.workspace_name,
                aml_auth=self.config.aml.auth,
                aml_tenant=self.config.aml.tenant,
                aml_force=self.config.aml.force,
            )  # NOTE: this also stores aml workspace in internal global variable

    def build_and_submit_new_pipeline(self):
        log.info(f"Building Pipeline [{self.__class__.__name__}]...")
        pipeline_function = self.build(self.config)

        log.info("Creating Pipeline Instance...")
        pipeline = self.pipeline_instance(pipeline_function, self.config)

        log.info("Overriding compute target if upstream steps use HDI DC...")
        pipeline = self._update_hdi_dc_configs(pipeline)

        log.info("Validating...")
        pipeline.validate(workspace=current_workspace())

        if self.config.run.export:
            log.info(f"Exporting to {self.config.run.export}...")
            with open(self.config.run.export, "w") as export_file:
                export_file.write(
                    pipeline._get_graph_json(workspace=current_workspace())
                )

        if self.config.run.publish:
            self._publish_to_endpoint(pipeline)

        if self.config.run.submit:
            pipeline_tags = self._parse_pipeline_tags()
            pipeline_tags.update({"shrike": __version__})
            pipeline_tags.update(self.repository_info)
            log.info(f"Submitting Experiment... [tags={pipeline_tags}]")

            # pipeline_run is of the class "azure.ml.component.run", which
            # is different from "azureml.pipeline.core.PipelineRun"
            pipeline_run = pipeline.submit(
                workspace=current_workspace(),
                experiment_name=self.config.run.experiment_name,
                description=self.config.run.experiment_description,
                display_name=self.config.run.display_name,
                tags=pipeline_tags,
                default_compute_target=self.config.compute.default_compute_target,
                regenerate_outputs=self.config.run.regenerate_outputs,
                continue_on_step_failure=self.config.run.continue_on_failure,
            )

            # Forece pipeline_run to be of the class "azureml.pipeline.core.PipelineRun"
            pipeline_run = PipelineRun(
                experiment=pipeline_run._experiment,
                run_id=pipeline_run._id,
            )
            return pipeline_run

        else:
            log.info(
                "Exiting now, if you want to submit please override run.submit=True"
            )
            self.__class__.BUILT_PIPELINE = (
                pipeline  # return so we can have some unit tests done
            )
            return

    def run(self):
        """Run pipeline using arguments"""
        # set logging level
        if self.config.run.verbose:
            logging.getLogger().setLevel(logging.DEBUG)

        # Log the telemetry information in the Azure Application Insights
        try:
            telemetry_logger = TelemetryLogger(
                enable_telemetry=not self.config.run.disable_telemetry
            )
            telemetry_logger.log_trace(
                message=f"shrike.pipeline=={__version__}",
                properties={"custom_dimensions": {"configuration": str(self.config)}},
            )
        except Exception as ex:
            log.debug(
                f"Sending trace log messages to application insight is not successful. The exception message: {ex}"
            )

        # Check whether the experiment name is valid
        self.validate_experiment_name(self.config.run.experiment_name)

        # Check whether the workspace name, resource group name, and subscription id are valid
        self.validate_connection_attribute(
            self.config.aml.workspace_name, "workspace"
        )  # self.config.aml.workspace_name
        self.validate_connection_attribute(
            self.config.aml.resource_group, "resource_group"
        )
        self.validate_connection_attribute(
            self.config.aml.subscription_id, "subscription_id"
        )

        self.repository_info = get_repo_info()
        log.info(f"Running from repository: {self.repository_info}")

        log.info(f"azureml.core.VERSION = {azureml.core.VERSION}")
        self.connect()

        pipeline_run = None
        if self.config.run.resume:
            if not self.config.run.pipeline_run_id:
                raise Exception(
                    "To be able to use --resume you need to provide both --experiment-name and --run-id."
                )

            log.info(f"Resuming Experiment {self.config.run.experiment_name}...")
            experiment = Experiment(
                current_workspace(), self.config.run.experiment_name
            )
            log.info(f"Resuming PipelineRun {self.config.run.pipeline_run_id}...")
            # pipeline_run is of the class "azureml.pipeline.core.PipelineRun"
            pipeline_run = PipelineRun(experiment, self.config.run.pipeline_run_id)
        else:
            keep_modified_files, override = False, False
            yaml_to_be_recovered = []

            if self.config.tenant_overrides.allow_override:
                log.info("Check if tenant is consistent with spec yaml")
                override, mapping = self._check_if_spec_yaml_override_is_needed()
                if override:
                    try:
                        tenant = self.config.aml.tenant
                        log.info(
                            f"Performing spec yaml override to adapt to tenant: {tenant}."
                        )
                        yaml_to_be_recovered = self._override_spec_yaml(mapping)
                        keep_modified_files = (
                            self.config.tenant_overrides.keep_modified_files
                        )
                    except BaseException as e:
                        log.error(f"An error occurred, override is not successful: {e}")
                        raise e
            try:
                pipeline_run = self.build_and_submit_new_pipeline()
            except BaseException as e:
                log.error(f"An error {e} occurred during pipeline submission.")
                if override:
                    log.info("Now trying to recover overrides.")
                self._recover_tenant_overrides(override, yaml_to_be_recovered, False)
                raise
            else:
                self._recover_tenant_overrides(
                    override, yaml_to_be_recovered, keep_modified_files
                )

        if not pipeline_run:
            # not submitting code, exit now
            return
        # launch the pipeline execution
        log.info(f"Pipeline Run Id: {pipeline_run.id}")
        log.info(
            f"""
#################################
#################################
#################################

Follow link below to access your pipeline run directly:
-------------------------------------------------------

{pipeline_run.get_portal_url()}

#################################
#################################
#################################
        """
        )

        if self.config.run.canary:
            log.info(
                "*** CANARY MODE ***\n----------------------------------------------------------"
            )
            pipeline_run.wait_for_completion(show_output=True)

            # azureml.pipeline.core.PipelineRun.get_status(): ["Running", "Finished", "Failed"]
            # azureml.core.run.get_status(): ["Running", "Completed", "Failed"]
            if pipeline_run.get_status() in ["Finished", "Completed"]:
                log.info("*** PIPELINE FINISHED, TESTING WITH canary() METHOD ***")
                self.canary(self.config, pipeline_run.experiment, pipeline_run)
                log.info("OK")
            elif pipeline_run.get_status() == "Failed":
                log.info("*** PIPELINE FAILED ***")
                raise Exception("Pipeline failed.")
            else:
                log.info("*** PIPELINE STATUS {} UNKNOWN ***")
                raise Exception("Pipeline status is unknown.")

        else:
            if not self.config.run.silent:
                webbrowser.open(url=pipeline_run.get_portal_url())

            # This will wait for the completion of the pipeline execution
            # and show the full logs in the meantime
            if self.config.run.resume or self.config.run.wait:
                log.info(
                    "Below are the raw debug logs from your pipeline execution:\n----------------------------------------------------------"
                )
                pipeline_run.wait_for_completion(show_output=True)

    @classmethod
    def main(cls):
        """Pipeline helper main function, parses arguments and run pipeline."""
        config_dict = cls._default_config()

        @hydra.main(config_name="default")
        def hydra_run(cfg: DictConfig):
            # merge cli config with default config
            cfg = OmegaConf.merge(config_dict, cfg)

            arg_parser = argparse.ArgumentParser()
            arg_parser.add_argument("--config-dir")
            args, _ = arg_parser.parse_known_args()
            cfg.run.config_dir = os.path.join(
                HydraConfig.get().runtime.cwd, args.config_dir
            )

            log.info("*** CONFIGURATION ***")
            log.info(OmegaConf.to_yaml(cfg))

            # create class instance
            main_instance = cls(cfg)

            # run
            main_instance.run()

        hydra_run()

        return cls.BUILT_PIPELINE  # return so we can have some unit tests done



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
PyTest suite for testing if run.py is aligned with module specification:
"""
import os
import sys
from pathlib import Path
import traceback
import argparse
from collections import namedtuple
import warnings
import pytest

from ruamel import yaml
from azure.ml.component._core._component_definition import CommandComponentDefinition

from shrike.pipeline.testing.importer import (
    import_and_test_class,
    dynamic_import_module,
)
from shrike.pipeline.testing.components import (
    component_spec_yaml_exists_and_is_parsable,
    component_uses_private_acr,
    component_uses_private_python_feed,
    component_run_py_import,
    component_run_get_arg_parser,
    if_arguments_from_component_spec_match_script_argparse,
)


def module_spec_yaml_exists_and_is_parsable(module):
    """Tests the presence of module specifications in yaml (and return it)"""
    return component_spec_yaml_exists_and_is_parsable(
        os.path.join(module, "module_spec.yaml")
    )


def module_uses_private_acr(module, acr_url):
    """Tests base image in private ACR"""
    component_uses_private_acr(os.path.join(module, "module_spec.yaml"), acr_url)


def module_uses_private_python_feed(module, feed_url):
    """Tests private python feed referenced in conda"""
    component_uses_private_python_feed(
        os.path.join(module, "module_spec.yaml"), feed_url
    )


def run_py_import(module):
    """Try importing run.py, just to check if basic script passes syntax/imports checks"""
    component_run_py_import(os.path.join(module, "module_spec.yaml"))


def module_run_get_arg_parser(module):
    """Tests is module run.py has function get_arg_parser(parser)"""
    component_run_get_arg_parser(os.path.join(module, "module_spec.yaml"))


def if_arguments_from_module_spec_match_script_argparse(module):
    """Tests alignment between module_spec arguments and script parser arguments"""
    if_arguments_from_component_spec_match_script_argparse(
        os.path.join(module, "module_spec.yaml")
    )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Python utilities to aid "compliant experimentation" - training
machine learning models without seeing the training data.
"""

__version__ = "1.22.0"



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Namespace containing Spark .NET utilities.

Inspired by the closed-source implementations:

- https://dev.azure.com/msdata/Vienna/_git/aml-ds?version=GC94d20cb3f190e942b016c548308becc107fcede8&path=/recipes/signed-components/canary-hdi/run.py  # noqa: E501
- https://dev.azure.com/eemo/TEE/_git/TEEGit?version=GC8f000c8c61ae67cf1009d7a70753f0175968ef81&path=/Offline/SGI/src/python/src/sparknet/run.py  # noqa: E501
- https://o365exchange.visualstudio.com/O365%20Core/_git/EuclidOdinML?version=GC849128898f37d138725695fbba3992cd5f5f4474&path=/sources/dev/Projects/dotnet/OdinMLDotNet/DotnetSpark-0.4.0.py  # noqa: E501
"""


import argparse
import logging
from shrike.compliant_logging import DataCategory, enable_compliant_logging
from shrike.compliant_logging.exceptions import (
    PublicRuntimeError,
    PublicValueError,
    PublicArgumentError,
)
import sys
from typing import Any, List, Optional
import uuid


log = logging.getLogger(__name__)


try:
    import py4j
    from pyspark.sql import SparkSession
    from pyspark import SparkContext
except ImportError as e:
    log.warning(f"Unable to import py4j or pyspark: {e}")
    py4j = Any
    SparkSession = Any
    SparkContext = Any


def get_default_spark_session() -> SparkSession:
    """
    Resolve a default Spark session for running Spark .NET applications.
    """
    # https://stackoverflow.com/a/534847
    random = str(uuid.uuid4())[:8]
    app_name = f"spark-net-{random}"
    log.info(f"Application name: {app_name}", category=DataCategory.PUBLIC)
    rv = SparkSession.builder.appName(app_name).getOrCreate()
    return rv


def full_pyfile_path(spark: SparkSession, py_file_name: str) -> str:
    """
    Resolve the full HDFS path of a file out of the Spark Session's PyFiles
    object.
    """
    spark_conf = spark.sparkContext.getConf()
    py_files = spark_conf.get("spark.yarn.dist.pyFiles")
    file_names = py_files.split(",")

    log.info(
        f"Searching py_files for {py_file_name}: '{py_files}'",
        category=DataCategory.PUBLIC,
    )

    for file_name in file_names:
        if file_name.split("/")[-1] == py_file_name:
            return file_name

    raise PublicValueError(f"py_files do not contain {py_file_name}: {py_files}")


def java_args(spark: SparkSession, args: List[str]):
    """
    Convert a Python list into the corresponding Java argument array.
    """
    rv = SparkContext._gateway.new_array(spark._jvm.java.lang.String, len(args))

    # https://stackoverflow.com/a/522578
    for index, arg in enumerate(args):
        rv[index] = arg

    return rv


def run_spark_net_from_known_assembly(
    spark: SparkSession, zip_file_name: str, assembly_name: str, args: List[str]
) -> None:
    """
    Invoke the binary `assembly_name` inside `zip_file_name` with the command
    line parameters `args`, using the provided Spark session. Print the Java
    stack trace if the job fails.
    """
    fully_resolved_zip_file_name = full_pyfile_path(spark, zip_file_name)
    dotnet_args = [fully_resolved_zip_file_name, assembly_name] + args

    log.info(
        f"Calling dotnet with arguments: {dotnet_args}", category=DataCategory.PUBLIC
    )
    dotnet_args_java = java_args(spark, dotnet_args)

    message = None

    try:
        spark._jvm.org.apache.spark.deploy.dotnet.DotnetRunner.main(dotnet_args_java)
    except py4j.protocol.Py4JJavaError as err:
        log.error("Dotnet failed", category=DataCategory.PUBLIC)
        for line in err.java_exception.getStackTrace():
            log.error(str(line), category=DataCategory.PUBLIC)

        message = f"{err.errmsg} {err.java_exception}"

    if message:
        # Don't re-raise the existing exception since it's unprintable.
        # https://github.com/bartdag/py4j/issues/306
        raise PublicRuntimeError(message)

    log.info("Done running dotnet", category=DataCategory.PUBLIC)


def run_spark_net(
    zip_file: str = "--zipFile",
    binary_name: str = "--binaryName",
    spark: Optional[SparkSession] = None,
    args: Optional[list] = None,
) -> None:
    """
    Easy entry point to one-line run a Spark .NET application. Simplest sample
    usage is:

    > run_spark_net_with_smart_args()
    """

    if not spark:
        spark = get_default_spark_session()

    if not args:
        args = sys.argv

    enable_compliant_logging()

    parser = argparse.ArgumentParser()
    parser.add_argument(zip_file, dest="ZIP_FILE")
    parser.add_argument(binary_name, dest="BINARY_NAME")

    try:
        (known_args, unknown_args) = parser.parse_known_args(args)

        zf = known_args.ZIP_FILE
        bn = known_args.BINARY_NAME
    except BaseException as e:
        raise PublicArgumentError(None, str(e)) from e

    run_spark_net_from_known_assembly(spark, zf, bn, unknown_args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Compliant logging utilities.
"""

from .constants import DataCategory  # noqa: F401

from .data_conversions import (  # noqa: F401
    is_numpy_array,
    is_pandas_dataframe,
    is_pandas_series,
    is_spark_dataframe,
    is_vaex_dataframe,
    get_numpy_array_info,
    get_pandas_dataframe_info,
    get_pandas_series_info,
    get_spark_dataframe_info,
    get_vaex_dataframe_info,
    collect_spark_dataframe,
    collect_pandas_dataframe,
    collect_vaex_dataframe,
    pandas_dataframe_schema,
    pandas_series_to_list,
    numpy_array_to_list,
    spark_dataframe_schema,
    vaex_dataframe_schema,
)

from .logging import (  # noqa: F401
    enable_compliant_logging,
    enable_confidential_logging,
    is_eyesoff,
)
from .system_info import provide_system_info  # noqa: F401
from .exceptions import prefix_stack_trace  # noqa: F401



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Provides a list of conversion functions based
on which libraries are available for import
"""

from typing import Any, Optional
from .exceptions import PublicRuntimeError


# spark functions
def is_spark_dataframe(obj: Any) -> bool:  # type: ignore
    """
    Checks if the given object is a spark dataframe.

    Args:
        obj (Any): The object to check

    Returns:
        bool: True if a dataframe, otherwise False
    """
    try:
        from pyspark.sql import DataFrame  # noqa
    except Exception:
        return False
    return isinstance(obj, DataFrame)


def get_spark_dataframe_info(df: Any) -> str:  # type: ignore
    """
    Provides information about the spark dataframe

    Args:
        obj (Any): The dataframe to provide information for

    Returns:
        str: info string about the dataframe
    """
    try:
        from pyspark.sql import DataFrame  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Spark DataFrame is not supported in current environment"
        )

    try:
        # df.count() may fail if JAVA version is not expected
        return "Spark DataFrame (Row Count: {} / Column Count: {})".format(
            df.count(), len(df.columns)
        )
    except Exception:
        return "Failed to extract Spark DataFrame info"


def spark_dataframe_schema(df: Any, schema_map: dict = None) -> dict:  # type: ignore
    """
    Retrieves the database schema in a dict as format `name: datatype` (both str)

    Args:
        df (Any): DataFrame to analyze
        schema_map (dict, optional): Optional renaming of column names.
            Defaults to None.

    Returns:
        dict: Str to Str dict that contains column names and datatypes
    """
    try:
        from pyspark.sql import DataFrame  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Spark DataFrame is not supported in current environment"
        )

    if schema_map:
        return dict([(schema_map.get(n, n), str(t)) for n, t in df.dtypes])
    else:
        return dict([(n, str(t)) for n, t in df.dtypes])


def collect_spark_dataframe(df: Any) -> Optional[dict]:  # type: ignore
    """
    Collects the spark dataframe as a dictionary,
    where keys are column names and values are a list of values

    Args:
        df (Any): The dataframe to collect

    Returns:
        dict: The dict with column names as keys and list of column values.
            Returns None if conversion fails
    """
    try:
        from pyspark.sql import DataFrame  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Spark DataFrame is not supported in current environment"
        )

    try:
        return df.toPandas().to_dict("list")
    except Exception:
        return None


# vaex functions


def is_vaex_dataframe(obj: Any) -> bool:  # type: ignore
    """
    Checks if the given object is a vaex dataframe

    Args:
        obj (Any): Object to check

    Returns:
        bool: True if a DataFrame otherwise False
    """
    try:
        from vaex.dataframe import DataFrame as VaexDataFrame  # noqa
    except Exception:
        return False
    return isinstance(obj, VaexDataFrame)


def get_vaex_dataframe_info(df: Any) -> str:  # type: ignore
    """
    Retrieves a general info string to the given Vaex DataFrame

    Args:
        df (Any): DataFrame to check for

    Returns:
        str: Info string
    """
    try:
        from vaex.dataframe import DataFrame as VaexDataFrame  # noqa
    except Exception:
        raise PublicRuntimeError("Vaex DataFrame not supported in current environment")

    return "Vaex DataFrame (Row Count: {} / Column Count: {})".format(
        df.count(), len(df.get_column_names())
    )


def vaex_dataframe_schema(  # type: ignore
    df: Any, schema_map: dict = None
) -> dict:  # type: ignore
    """
    Retrieves the schema of the given DataFrame

    Args:
        df (Any): DataFrame to provide the schema for
        schema_map (dict, optional): Mapping of the schema to

    Returns:
        dict: str to str dict that contains column names and datatypes
    """
    try:
        from vaex.dataframe import DataFrame as VaexDataFrame  # noqa
    except Exception:
        raise PublicRuntimeError("Vaex DataFrame not supported in current environment")

    if schema_map:
        return {
            schema_map.get(col, col): df[col].dtype.name
            for col in df.get_column_names()
        }
    else:
        return {
            column_name: df.data_type(column_name)
            for column_name in df.get_column_names()
        }


def collect_vaex_dataframe(df: Any) -> Optional[dict]:  # type: ignore
    """
    Collects the vaex dataframe as a dictionary,
    where keys are column names and values are a list of values

    Args:
        df (Any): The dataframe to collect

    Returns:
        dict: The dictionary with column names as keys and list of column values.
            Returns None if conversion fails
    """
    try:
        from vaex.dataframe import DataFrame as VaexDataFrame  # noqa
    except Exception:
        raise PublicRuntimeError("Vaex DataFrame not supported in current environment")

    try:
        return df.to_pandas_df().to_dict("list")
    except Exception:
        return None


# numpy functions
def is_numpy_array(obj: Any) -> bool:  # type: ignore
    """
    Checks if the given object is a numpy array

    Args:
        obj (Any): The object to check

    Returns:
        bool: True if the object is a numpy array otherwise False
    """
    try:
        import numpy as np  # noqa
    except Exception:
        return False
    return isinstance(obj, np.ndarray)


def get_numpy_array_info(arr: Any) -> str:  # type: ignore
    """
    Retrieves an info string from the given numpy array

    Args:
        arr (Any): The array to generate string from

    Returns:
        str: Info String
    """
    try:
        import numpy as np  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Numpy Array not supported in the current environment."
        )
    return "Numpy Array (Shape: {})".format(arr.shape)


def numpy_array_to_list(arr: Any) -> list:  # type: ignore
    """
    Converts the given numpy array into a python list

    Args:
        arr (Any): Array to convert

    Returns:
        list: Converted (potentially multi-dimensional) list
    """
    try:
        import numpy as np  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Numpy Array not supported in the current environment."
        )
    return arr.tolist()


# pandas functions
def is_pandas_series(obj: Any) -> bool:  # type: ignore
    """
    Checks if the given object is a pandas series

    Args:
        obj (Any): Object to check

    Returns:
        bool: True if object is a series otherwise False
    """
    try:
        import pandas as pd  # noqa
    except Exception:
        return False
    return isinstance(obj, pd.Series)


def is_pandas_dataframe(obj: Any) -> bool:  # type: ignore
    """
    Checks if the given object is a pandas DataFrame

    Args:
        obj (Any): Object to check

    Returns:
        bool: True if object is a DataFrame otherwise False
    """
    try:
        import pandas as pd  # noqa
    except Exception:
        return False
    return isinstance(obj, pd.DataFrame)


def get_pandas_series_info(series: Any) -> str:  # type: ignore
    """
    Retrieves an info string from the given pandas series

    Args:
        series (Any): Series to convert

    Returns:
        str: Info String containing the Row Count
    """
    try:
        import pandas as pd  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Pandas Series not supported in the current environment."
        )
    return "Pandas Series (Row Count: {})".format(series.count())


def get_pandas_dataframe_info(df: Any) -> str:  # type: ignore
    """
    Retrieves an info string from the given pandas DataFrame

    Args:
        df (Any): DataFrame to convert

    Returns:
        str: Info String containing Row and Column Count
    """
    try:
        import pandas as pd  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Pandas Series not supported in the current environment."
        )
    return "Pandas DataFrame (Row Count: {} / Column Count: {})".format(
        df.shape[0], df.shape[1]
    )


def pandas_series_to_list(series: Any) -> list:  # type: ignore
    """
    Converts the given Pandas Series to a list

    Args:
        series (Any): Series to convert

    Returns:
        list: List output
    """
    try:
        import pandas as pd  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Pandas Series not supported in the current environment."
        )
    return series.tolist()


def pandas_dataframe_schema(df: Any) -> dict:  # type: ignore
    """
    Retrieves the schema of the dataframe as dict in name to datatype format

    Args:
        df (Any): DataFrame to convert

    Returns:
        dict: str-to-str Dict that contains column names and datatypes
    """
    try:
        import pandas as pd  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Pandas Series not supported in the current environment."
        )
    return dict([(col, df[col].dtype.name) for col in df.columns])


def collect_pandas_dataframe(df: Any) -> dict:  # type: ignore
    """
    Converts the given pandas DataFrame to a list

    Args:
        df (Any): DataFrame to convert

    Returns:
        list: Converted Dataframe as multi-dimensional list
    """
    try:
        import pandas as pd  # noqa
    except Exception:
        raise PublicRuntimeError(
            "Pandas Series not supported in the current environment."
        )
    return df.to_dict("list")



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
APIs for federated learning.
"""
from shrike.pipeline.pipeline_helper import AMLPipelineHelper
from shrike.pipeline.module_helper import module_reference
from shrike._core import experimental
from coolname import generate_slug
from dataclasses import dataclass
from typing import List, Optional, Union
from azure.ml.component import dsl
from azure.ml.component.pipeline import Pipeline
from azure.ml.component.component import Component, Output
import logging
from omegaconf import OmegaConf, open_dict, DictConfig
from pathlib import Path
from typing import Callable, Dict
import collections.abc

log = logging.getLogger(__name__)

EXPERIMENTAL_WARNING_MSG = (
    "Federated APIs are experimental and could change at any time."
)


def dataset_name_to_reference(
    component_instance: Component, dataset_names: list
) -> Dict[str, Output]:
    res = {}
    if not dataset_names:
        return res
    for dataset in dataset_names:
        res[dataset] = component_instance.outputs[dataset]
    return res


@dataclass
class StepOutput:
    """Output object from preprocess/midprocess/postprocess/training step in a federated pipeline."""

    step: Union[Component, Pipeline] = None
    outputs: List[str] = None


class FederatedPipelineBase(AMLPipelineHelper):
    """
    Base class for Federated Learning pipelines.
    """

    def __init__(self, config, module_loader=None):
        super().__init__(config, module_loader=module_loader)

        package_path = Path(__file__).parent / "components"
        fedavg_component = DictConfig(
            module_reference(
                key="fl_fedavg_pytorch",
                name="fl_fedavg_pytorch",
                yaml=str(package_path) + "/fedavg/spec.yaml",
            )
        )
        self.config.modules.manifest.append(fedavg_component)
        self.module_loader.modules_manifest = OmegaConf.merge(
            self.module_loader.modules_manifest, {"fl_fedavg_pytorch": fedavg_component}
        )
        self.cur_fl_iteration: Optional[int] = None
        self.output_prefix = "fl_output_"

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def create_base_name(self) -> str:
        """The training outputs from each silo will be stored at <base_name>/<silo_name> on the central storage.
        By default, it will return a name of 2 word. User can override this method.

        Returns:
            base_name
        """
        rv = "fl-"
        rv += generate_slug(2)
        return rv

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def create_iteration_name(self) -> str:
        """Display name for each iteration's subgraph, containing letters, numbers, and underscores only. The default value is "iteration n"."""
        return f"iteration_{self.cur_fl_iteration}"

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def create_silo_specific_name(self, silo_name) -> str:
        """Display name for a silo's training in iteration n's subgraph, containing letters, numbers, and underscores only. The default value is "silo_<silo_name>__iteration_n"."""
        return f"silo_{silo_name}__iteration_{self.cur_fl_iteration}"

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def create_subgraph_name(self) -> str:
        """Display name for each subgraph"""
        return "Federated Learning Subgraph"

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def preprocess(self, config: DictConfig, **input) -> StepOutput:
        """
        Optional user-defined preprocess step. The outputs will be distributed to each silo's datastore.

        Returns:
            a component/subgraph instance, and a list of output dataset names to be passed to the downstream pipeline.
        """
        pass

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def midprocess(self, config: DictConfig, **input) -> StepOutput:
        """
        User-defined midprocess step which reads outputs from `train` in each silo. The outputs will be distributed to each silo's datastore.
        By default, it calls the built-in FedAvg component (un-signed).

        Returns:
            a component/subgraph instance, and a list of output dataset name to be passed to the downstream pipeline.
        """
        fedavg_component = self.component_load("fl_fedavg_pytorch")
        num_models = len(self.config.federated_config.silos)
        if len(input) != num_models:
            log.error(
                "The input model weights are inconsistent with the number of models."
            )
            raise Exception
        if "agg_weights" in self.config.federated_config.params:
            weights = self.config.federated_config.params.agg_weights
        else:
            weights = ",".join([str(1 / num_models)] * num_models)
        model_name = (
            self.config.federated_config.params.model_name
            if "model_name" in self.config.federated_config.params
            else None
        )
        updated_input = {}
        for k, v in input.items():
            idx = k[k.index("silo") + 4 : k.index("_")]
            updated_input["model_input_dir_" + idx] = v
        fedavg_step = fedavg_component(
            **updated_input,
            num_models=num_models,
            weights=weights,
            model_name=model_name,
        )
        return StepOutput(fedavg_step, ["model_output_dir"])

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def postprocess(self, config: DictConfig, **kwargs) -> StepOutput:
        """
        Optional user-defined postprocess step which reads outputs from the final `midprocess` step and writes to the `noncompliant_datastore`.

        Returns:
            a component/subgraph instance, and a list of output dataset names.
        """
        pass

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def train(
        self, config: DictConfig, silo: DictConfig, input: Output = None, **kwargs
    ) -> StepOutput:
        """
        User-defined train step happening at each silo. This reads outputs from `preprocess` or `midprocess`, and sends outputs back to the `noncompliant_datastore`.

        Returns:
            a component/subgraph instance, and a list of output dataset names to be passed to the downstream pipeline.
        """
        pass

    def _data_transfer(
        self, input_list: dict, destination_datastore: str, base_name: str
    ) -> dict:
        """Moves data `input_list` to `destination_datastore` with path prefixed by `base_name`."""
        # assume one shared ADF
        dts_name = self.config.federated_config.data_transfer_component
        try:
            data_transfer_component = self.component_load(
                dts_name  # this component is subject to change
            )
            log.info(f"Using {dts_name} as data transfer component.")
        except Exception as e:
            log.error(
                f"Please specify the data transfer component name registered in your workspace in module manifest and `federated_config.data_transfer_component` pipeline yaml."
            )
            raise (e)

        res = {}
        if not input_list:
            return res
        for name, input in input_list.items():
            data_transfer_step = data_transfer_component(source_data=input)
            self.apply_smart_runsettings(data_transfer_step, datatransfer=True)
            data_transfer_step.outputs.destination_data.configure(
                datastore=destination_datastore,
                path_on_datastore=base_name + "/" + name,
            )
            res[name] = data_transfer_step.outputs.destination_data
        return res

    def _update_nested_dict_with_non_none_values(self, dict1, dict2):
        """Update dict1 with non-none values in dict2"""
        for k, v in dict2.items():
            if not v:
                continue
            if isinstance(v, collections.abc.Mapping):
                dict1[k] = self._update_nested_dict_with_non_none_values(
                    dict1.get(k, {}), v
                )
            else:
                log.info(f"Setting {k} to {v}.")
                dict1[k] = v
        return dict1

    def _get_user_defined_output_configs(self, component_instance):
        output_configs = {}
        for output_key in component_instance.outputs:
            output_instance = getattr(component_instance.outputs, output_key)
            config = {
                "datastore": output_instance._datastore,
                "mode": output_instance._mode,
                "path_on_compute": output_instance._path_on_compute,
                "path_on_datastore": output_instance._dataset_output_options.path_on_datastore
                if output_instance._dataset_output_options
                else None,
            }
            for k, v in dict(config).items():
                if not v:
                    del config[k]
            if config:
                output_configs[output_instance._name] = config
        return output_configs

    def _apply_user_defined_output_configs(self, outputs, output_configs):
        for output_key in outputs:
            if output_key in output_configs:
                outputs[output_key].configure(**output_configs[output_key])
        return outputs

    def _process_at_orchestrator(self, prev: StepOutput) -> dict:
        """Processes outputs at the orchestrator."""
        step = prev.step
        if isinstance(step, Pipeline):
            prev_output = step.outputs
        elif isinstance(step, Component):
            prev_output = dataset_name_to_reference(step, prev.outputs)
            log.info(f"Checking user override runsettings for {step.name}...")
            output_configs = self._get_user_defined_output_configs(step)

            runsettings = step.runsettings._get_values()
            self.apply_smart_runsettings(step)
            merged_runsettings = self._update_nested_dict_with_non_none_values(
                step.runsettings._get_values(), runsettings
            )
            step.runsettings._set_section_by_dict(merged_runsettings)
            prev_output = self._apply_user_defined_output_configs(
                prev_output, output_configs
            )
        else:
            raise Exception(
                f"The output of preprocess/midprocess/postprocess step must be a Pipeline object or a Component object. You are using {type(prev.step)}."
            )
        return prev_output

    def _merge_config(self):
        """Merges and simplifies federated_config."""
        # override priority: default_config < customized shared config < per-silo setting
        # everything in customized shared config have equal priority
        federated_config = self.config.federated_config
        default_config = federated_config.config_group.get("default_config", {})
        for silo_name, silo_config in federated_config.silos.items():
            customized_config = {}
            if "inherit" in silo_config:
                for inherit_config in silo_config["inherit"]:
                    customized_config = OmegaConf.merge(
                        customized_config, federated_config.config_group[inherit_config]
                    )
            log.info(
                f"=== Original config for silo {silo_name}: {self.config.federated_config.silos[silo_name]}"
            )
            with open_dict(default_config):
                merge = OmegaConf.merge(default_config, customized_config)
            log.info(f"=== Shared config {merge}")
            with open_dict(self.config.federated_config.silos[silo_name]):
                self.config.federated_config.silos[silo_name] = OmegaConf.merge(
                    merge, silo_config
                )
                self.config.federated_config.silos[silo_name].pop("inherit", None)
            log.info(
                f"=== Merged config for silo {silo_name}: {self.config.federated_config.silos[silo_name]}"
            )

    def _check_if_data_transfer_is_activated_and_process(
        self, datastore_name, base_name, **inputs
    ):
        deactivate_data_transfer = self.config.federated_config.deactivate_data_transfer
        if deactivate_data_transfer:
            log.info("Data transfer is disabled; training outputs will remain in silo.")
            return inputs
        else:
            log.info(f"Moving data {inputs.keys()} into {datastore_name}.")
            return self._data_transfer(
                input_list=inputs,
                destination_datastore=datastore_name,
                base_name=base_name,
            )

    def _train_in_silo_once(
        self, silo_name: str, silo: DictConfig, cool_name: str, **prev_output
    ) -> dict:
        """Runs the "training" step once at one silo."""
        prev_output = self._check_if_data_transfer_is_activated_and_process(
            silo.datastore, cool_name, **prev_output
        )
        if len(prev_output) == 1:
            prev_output = {"input": list(prev_output.values())[0]}
        train_step = self.train(self.config, silo=silo, **prev_output)

        step = train_step.step
        if isinstance(step, Pipeline):
            train_output = step.outputs
            for node in step._expand_pipeline_nodes():
                node._node_name = self.create_silo_specific_name(silo_name)
        elif isinstance(step, Component):
            train_output = dataset_name_to_reference(step, train_step.outputs)
            log.info(
                f"Checking user override runsettings for {silo_name} / {step.name}..."
            )
            output_configs = self._get_user_defined_output_configs(step)

            runsettings = step.runsettings._get_values()
            self.apply_smart_runsettings(
                step,
                target=silo.compute,
                datastore_name=silo.datastore,
            )
            merged_runsettings = self._update_nested_dict_with_non_none_values(
                step.runsettings._get_values(), runsettings
            )
            step.runsettings._set_section_by_dict(merged_runsettings)
            train_output = self._apply_user_defined_output_configs(
                train_output, output_configs
            )

            step._node_name = self.create_silo_specific_name(silo_name)

        else:
            raise Exception(
                f"The output of train step must be a Pipeline object or a Component object. You are using {type(step)}"
            )

        silo_output = self._check_if_data_transfer_is_activated_and_process(
            self.config.compute.noncompliant_datastore,
            cool_name + "/" + silo_name,
            **train_output,
        )
        return silo_output

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def build(self, config: DictConfig):
        """Constructs the federated pipeline. User does not need to modify."""

        @dsl.pipeline(display_name=self.create_subgraph_name())
        def _subpipeline_function(**kwargs):
            cool_name = self.create_base_name()
            midprocess_input = {}
            i = 1
            for silo_name, silo in self.config.federated_config.silos.items():
                train_output = self._train_in_silo_once(
                    silo_name,
                    silo,
                    cool_name,
                    **kwargs,
                )
                if len(train_output) == 1:
                    # if there's only one output, forcibly change the name to "<silo_name>_input"
                    midprocess_input[f"silo{i}_input"] = list(train_output.values())[0]
                else:
                    for k, v in train_output.items():
                        midprocess_input[f"silo{i}_" + k] = v
                i += 1

            prev = self.midprocess(self.config, **midprocess_input)
            prev_output = self._process_at_orchestrator(prev) if prev else None
            output = {
                self.output_prefix + name: prev_output[name] for name in prev_output
            }
            return output

        @dsl.pipeline()
        def pipeline_function():
            self._merge_config()
            prev = self.preprocess(self.config)
            prev_output = self._process_at_orchestrator(prev) if prev else None

            for iter in range(self.config.federated_config.max_iterations):
                self.cur_fl_iteration = iter
                if prev_output:
                    prev_output = {
                        output_name.replace(self.output_prefix, ""): prev_output[
                            output_name
                        ]
                        for output_name in prev_output.keys()
                    }
                    subpipeline = _subpipeline_function(**prev_output)
                else:
                    subpipeline = _subpipeline_function()
                subpipeline.node_name = self.create_iteration_name()
                prev_output = subpipeline.outputs

            if len(prev_output) == 1:
                prev_output = {"input": list(prev_output.values())[0]}
            prev = self.postprocess(self.config, **prev_output)
            output = self._process_at_orchestrator(prev) if prev else None
            return output

        return pipeline_function

    def pipeline_instance(self, pipeline_function, config):
        """Creates an instance of the pipeline using arguments. User does not need to modify."""
        pipeline = pipeline_function()
        return pipeline
